require('colors')
const path = require('path')
const FormData = require('form-data')
const axios = require('axios')
const fs = require('fs')
const Queue = require('promise-queue-plus');

class CdnWebpackPlugin {
  constructor({ code, origin, api, timeout = 10000, retry = 0, concurrent = 1 }) {
    this.code = code
    this.api = api
    this.origin = origin
    this.timeout = timeout
    this.retry = retry
    this.concurrent = concurrent
    this.queue = Queue(this.concurrent, {
      'retry': this.retry,
      'retryIsJump': false,
      'timeout': this.timeout,
      'autoRun': true,
      'queueEnd': () => {
        console.log('Upload complete.'.magenta)
      },
      'workFinally': (queue) => {
        console.log(`(剩余上传数量：${queue.getLength()})`)
      }
    })
  }

  tasks({ assets, outputPath }) {
    return assets.reduce((tasks, file) => {
      if (!file.name.endsWith('.map') && !file.name.endsWith('.html')) {
        tasks.push(
          new Promise((resolve, reject) => {
            const form = new FormData()
            const filePaths = file.name.split('/')
            filePaths.pop()
            form.append('action', 'upload')
            form.append('type', 'aliyun')
            form.append('dir', `${this.code}/${filePaths.length > 0 ? filePaths.join('/') : ''}/`)
            form.append(
              'file',
              fs.createReadStream(path.resolve(outputPath, file.name))
            )
            axios
              .post(this.api, form, {
                headers: form.getHeaders()
              })
              .then(res => {
                if (res.data.code === 200) {
                  resolve(res)
                } else {
                  console.log(`资源上传失败: ${file.name}`.red)
                  reject(res)
                }
              })
              .catch(res => {
                console.log(`资源上传失败: ${file.name}`.red)
                reject(res)
              })
          })
        )
      }
      return tasks
    }, [])
  }

  pushQueue({ assets, outputPath }) {
    for (const file of assets) {
      if (!file.name.endsWith('.map') && !file.name.endsWith('.html') && !(file.size === 0)) {
        this.queue.add((resolve, reject) => {
          const form = new FormData()
          const filePaths = file.name.split('/')
          filePaths.pop()
          form.append('action', 'upload')
          form.append('type', 'aliyun')
          form.append('file_cover', 1)
          form.append('dir', `${this.code}/${filePaths.length > 0 ? filePaths.join('/') : ''}/`)
          form.append(
            'file',
            fs.createReadStream(path.resolve(outputPath, file.name))
          )
          axios
            .post(this.api, form, {
              headers: form.getHeaders(),
              timeout: this.timeout
            })
            .then(res => {
              if (res.data.code === 200) {
                console.log(`资源上传成功: ${file.name}`.green)
                resolve(res)
              } else {
                console.log(`资源上传异常: ${file.name}`.red)
                reject(res)
              }
            })
            .catch(res => {
              console.log(`资源上传失败: ${file.name}`.red)
              reject(res)
            })
        })
      }
    }
  }

  apply(compiler) {
    compiler.hooks.done.tap('CdnWebpackPlugin', stats => {
      console.log('\nUpload start.'.magenta)
      // Promise.all(this.tasks(stats.toJson())).then(() => {
      //   console.log('Upload complete.'.magenta)
      // })
      this.pushQueue(stats.toJson())
    })
  }
}

module.exports = CdnWebpackPlugin
